//
//  ndarray.h
//  RSvIDX_V4
//
//  Created by Ruben Ticehurst-James on 03/01/2023.
//

#ifndef ndarray_h
#define ndarray_h
#include "config.h"

struct ndarray_shape {
	ROW_COLUMN_TYPE rows, columns;
};

#endif /* ndarray_h */
